#!/usr/bin/env python3

import csv
import sys
import re
from bs4 import BeautifulSoup
from filters import filters
import logging
import json
from pathlib import Path
import os
import time
import codecs
import pandas as pd
from urllib.parse import urlparse


csv.field_size_limit(sys.maxsize)

log_file = "./logs/log_" + time.strftime("%Y%m%d-%H%M%S") + ".txt"
file_handler = logging.FileHandler(log_file)
console_handler = logging.StreamHandler()
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    handlers=[file_handler, console_handler])

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def clean_text(txt):
    txt = txt.replace('\n', " ")
    txt = re.sub("[ \t]+", " ", txt)
    return txt

def extract_html_table(html_table):
    table = []
    for tr in html_table.find_all(name="tr"):
        row = []
        for td in tr.find_all(name="td"):
            cols = 1
            if "colspan" in td.attrs:
                try:
                    cols = int(td["colspan"])
                except ValueError:
                    cols = 1
            rep = [clean_text(td.get_text().strip()) for _ in range(cols)]
            row = row + rep
        table.append(row)
    return table

def parse_text(bs, filter_dict, url):   #  we only need the url for logging purposes

    if "standard_filter" in filter_dict:
        standard_filter = filter_dict["standard_filter"]
        if not standard_filter:
            logger.debug("Skipping %s since not valid filter was provided.", url)
            return []

        logger.debug("Parsing %s", url)
        # get to the entry point
        ep = bs
        for entrypoint in standard_filter["entrypoints"]:
            attr = {}
            if "attr" in entrypoint:
                attr = {entrypoint["attr"]: entrypoint["value"]}
            ep = ep.find(name=entrypoint["tag"], attrs=attr)
            if not ep:
                logger.warning("Couldn't reach entrypoint %s having attribute %s for URL %s", entrypoint["tag"], attr, url)
                return []

        # retrieve meta tags
        meta_items = {} # extracted content of meta tags
        e = bs.find(name="meta", attrs={'property':'og:title'})
        if e and e.has_attr('content'):
            content = e['content']
            meta_items['og:title'] = content

        e = bs.find(name="meta", attrs={'property': 'og:description'})
        if e and e.has_attr('content'):
            content = e['content']
            meta_items['og:description'] = content

        e = bs.find(name="meta", attrs={'name': 'news_keywords'})
        if e and e.has_attr('content'):
            content = e['content']
            content = [c.strip() for c in content.split(',')]
            meta_items['news_keywords'] = content

        e = bs.find(name="meta", attrs={'name': 'keywords'})
        if e and e.has_attr('content'):
            content = e['content']
            content = [c.strip() for c in content.split(',')]
            meta_items['keywords'] = content

        # delete unwanted stuff
        if "delete" in standard_filter:
            for to_delete in standard_filter["delete"]:
                if "attrs" in to_delete:
                    for a in to_delete["attrs"]:
                        attr = {a["attr"]: a["values"]}
                        [unwanted.decompose() for unwanted in ep.find_all(name=to_delete["tag"], attrs=attr)]
                else:
                    [unwanted.decompose() for unwanted in ep.find_all(name=to_delete["tag"])]
        # extract tables
        html_tables = ep.find_all('table')
        tables = [extract_html_table(html_table) for html_table in html_tables]

        #delete tables
        [s.decompose() for s in ep.find_all('table')]

        # filter tags
        txt = ""
        if "tag_filters" in standard_filter:
            tags = [f["tag"] for f in standard_filter["tag_filters"]]
            txt = " ".join([tag.get_text(separator=" ").strip() for tag in ep.find_all(name=tags)])
        else:
            txt = ep.get_text(separator=" ").strip()
        txt = clean_text(txt)
        #result = {'text': txt, 'tables': tables}
        return [txt, json.dumps({'tables':tables}, ensure_ascii=False), json.dumps(meta_items, ensure_ascii=False)]

    elif filter_dict["custom_filter"]:
        result = filter_dict["custom_filter"](bs)
        if not result or result == "":
            logger.warning("Error when parsing %s", url)
        return result

    else:
        logger.warning("No valid filter type provided for %s", url)
        return []

def get_commandline_arguments():
    replace_existing_cleaned_files = False
    if len(sys.argv) > 2 and len(sys.argv) < 6:
        dirty_file_path = sys.argv[1]
        if not Path(dirty_file_path).is_file():
            print(dirty_file_path + " is not present.")
            return []
        dump_path = sys.argv[2]
        if not Path(dump_path).is_dir():
            print(dump_path + " is not present.")
            return []
        return [dirty_file_path, dump_path, ]
    else:
        print("Usage: python3 cleander_adv.py path_to_file /foo/bar/cleaned_dump ")
        return []

def feed_url_to_outlet(feed_url, outlet_frame):
    #outlet,outlet_url,country,feed_url,schedule
    filtered_outlet_frame = outlet_frame[outlet_frame["feed_url"] == feed_url]
    if filtered_outlet_frame.empty:
        return None
    else:
        return filtered_outlet_frame["outlet_url"].iloc[0]


def import_outlets(outlet_path):
    print("Importing outlets %s...", outlet_path)
    files = [os.path.join(outlet_path, f) for f in os.listdir(outlet_path) if os.path.isfile(os.path.join(outlet_path, f))]
    frames = [pd.read_csv(file,
                          names=["outlet", "outlet_url", "country", "feed_url", "schedule"])
              for file in files]
    outlet_frame = pd.concat(frames).drop_duplicates(subset=['feed_url'])
    return outlet_frame

#90´0987654321^  >BCV   BCBCXBCVBC
def clean_file(file_path, cleaned_dir):
    file_name = os.path.basename(file_path)          # TODO this is os dependent!
    cleaned_file_path = os.path.join(cleaned_dir, file_name + "-cleaned.csv")
    if Path(cleaned_file_path).is_file():   # only consider files that haven't been created yet!
        logger.info(cleaned_file_path + " already exists. Continuing...")
        return
    o = open(cleaned_file_path, 'w', encoding='utf-8')
    writer = csv.writer(o, quoting=csv.QUOTE_ALL)

    with open(file_path, 'rb') as csvfile:
        for line in csvfile:
            line = codecs.decode(line, 'utf-8')  # .encode('unicode_escape')
            reader = csv.DictReader([line], fieldnames=["timestamp", "feed_url", "article_url", "html"], delimiter=',', quotechar='"')
            row = list(reader)[0]

            #outlet_url = feed_url_to_outlet(row["feed_url"], outlets_frame)
            #domain = urlparse(outlet_url).netloc
            domain = urlparse(row["feed_url"]).netloc


            raw = row["html"].encode('utf-8').decode('unicode_escape')  # .encode('utf-8').decode('unicode_escape')

            #raw = raw.replace('★★★', "")\
            #     .replace("\\xc3\\xa4", "ä")\
            #     .replace("\\xe2\\x80\\x94", " - ")\
            #     .replace("\\xe2\\x80\\x9d","\"")\
            #     .replace("\\xc2", " ")\
            #     .replace("\\xc3\\xb6", "ö")\
            #     .replace("\\xc3\\x96", "Ö")\
            #     .replace("\\xc3\\xbc", "ü")\
            #     .replace("\\xe2\\x80\\x93", " - ")\
            #     .replace("\\xa3", "£")\
            #     .replace("\\xa0", " ")\
            #     .replace("\\n", "\n ")\
            #     .replace("\\xe4", "ä")\
            #     .replace("\\xfc", "ü")\
            #     .replace("\\xf6", "ö")\
            #     .replace("\\xdf", "ß") \
            #     .replace("\\xe9", "é") \
            #     .replace("\\xbd", " 0.5") \
            #     .replace("\\xbb", " >> ") \
            #     .replace("\\xed", "i") \
            #     .replace("\\t", " ")\
            #     .replace("\\r", "")\
                #.replace("\\\\", "\\")
                #\
                #.replace("\u00d6","Ö")

            #raw = raw.replace("\\xa0", " ").replace("\\n", "\n ").replace("\\xe4", "ä").replace("\\xfc", "ü").replace(
            #    "\\xf6", "ö").replace("\\xdf", "ß").replace("\\t", " ").replace("\\r", "")
            if not raw:
                logger.debug("row %s with url %s was empty.", row["timestamp"], row["URL"])
                continue
            if domain not in filters:
                logger.warning("Error parsing %s. Domain %s does not have a filter!", file_path, domain)
                continue
            bs = BeautifulSoup(raw, "lxml")
            filter = filters[domain]
            if not filter:
                continue
            result = parse_text(bs, filter, row["article_url"]) # [text, tables]
            if not result:
                # This is an error but all necessary log output was already written in parse_text(.)
                continue
            title = bs.title.string.strip()
            #print(result[0])
            save_row = [row["timestamp"], domain, row["feed_url"], row["article_url"], title, result[0], result[1], result[2]]
            writer.writerow(save_row)
    o.close()


def run():
    args = get_commandline_arguments()
    if not args:
        sys.exit(os.EX_USAGE)
    else:
        #outlets_frame = import_outlets(args[2])

        #clean_file(args[0], args[1], outlets_frame)
        clean_file(args[0], args[1])


if __name__ == "__main__":
    run()
